################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from py_fast_trie.x_fast import XFastTrie as XFastTrie  # noqa
from py_fast_trie.y_fast import YFastTrie as YFastTrie  # noqa
